#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "ees_eics.h"
#include "config.h"
#include "xacml.h"
#include "xacml_server.h"
#include "eef_library.h"
#include "cgul/log_to_file/log_to_file.h"
#include "cgul_patched/network/net_common.h"

#define _MULTI_THREADED
#define PORT 9100
#define BACKLOG 100

char* config_file_path;                     /* config file path */
xacml_server_t server;                      /* xacml server object */
int signal_is_being_handled = 0;            /* signal handler state/lock */

void signal_handler(int sig);               /* main signal handler */
EES_RC process_xacml(void);                 /* creates and starts new xacml server instance */
EES_RC ees_start(void);                     /* initializes EEF */
EES_RC ees_stop(void);                      /* terminates EEF */
EES_RC ees_loop(void);                      /* main loop - calls process_xacml and pauses */
EES_RC add_dummy_aos_obligation(void);      /* adds some stuff to the AOS */



/**
 * Adds some dummy uid/gid obligations in the AOS
 */
EES_RC add_dummy_aos_obligation(void){
	aos_context_t*   _context;
	aos_attribute_t *_uid, *_gid;

	char* _uid_str = "1234";
	char* _gid_str = "1234";

  /*fill aos */
  _uid = createAttribute();
  _gid = createAttribute();

  setAttributeId(_uid, "posix-uid");
  setAttributeValue(_uid, _uid_str, strlen(_uid_str));
  setAttributeId(_gid, "posix-gid");
  setAttributeValue(_gid, _gid_str, strlen(_gid_str));

  if((_context = createContext(OBLIGATION)) != NULL){
    setContextObligationId(_context, "uidgid");

    addAttribute(_context, _uid);
    addAttribute(_context, _gid);

    addContext(_context);
  }
  return EES_SUCCESS;
}

/**
 * Prepares logging, initializes EEF and puts the EEF in threading mode
 */
EES_RC ees_start(void){
  /* set log options */
  if(cgul_openlog("/tmp/ees_log") != 0){
    return EES_FAILURE;
  }
  cgul_set_ident(BINARY_NAME);
  cgul_set_log_options(CGUL_SYSLOG | CGUL_LOGFILE | CGUL_STDERR);

	/* reuse this part when implementing SIGHUP */
	cgul_log(LOG_INFO, "Parsing config file %s\n", config_file_path);
	if(EEF_init(config_file_path, cgul_log) == EES_FAILURE){
    /* at this point, EEF_log can/will be a reference to the second argument passed to EEF_init */
		EEF_log(LOG_ERR, "Failed to parse config file (%s), exiting...", config_file_path);
    return EES_FAILURE;
	} else {
		EEF_startThreading();
	}
  EEF_log(LOG_NOTICE, "EES Ready.\n");  
  return EES_SUCCESS;
}

/**
 * Main EES loop. Starts xacml server and pauses.
 */
EES_RC ees_loop(void){
  while(1){ /* this is to prevent falling out of the loop when a signal gets handled */
    signal_is_being_handled = 0;
    if(process_xacml() == EES_SUCCESS){
      printf("Paused.\n");
      pause();
    } else {
      EEF_log(LOG_ERR, "Unable to start xacml server\n");
      ees_stop();
      exit(EXIT_FAILURE);
    }
  }
  return EES_SUCCESS;
}

/**
 * Terminates EES
 */
EES_RC ees_stop(void){
  xacml_server_destroy(server);
  AOS_Clean();
  if(EEF_term() == EES_SUCCESS){
    EEF_log(LOG_NOTICE, "EEF terminated!\n");
  } else {
    EEF_log(LOG_NOTICE, "Error while terminating EEF!\n");
  }
  fflush(stdout);
  fflush(stderr);
  EEF_log(LOG_NOTICE, "EES Stopped.\n");  
  cgul_closelog();
  return EES_SUCCESS;
}

/**
 * Creates and starts xacml server.
 */
EES_RC process_xacml(void){
  int fd;
  /*shutdown(fd, 0);*/
  if((fd = createAndSetUpATCPServerSocket(PORT, BACKLOG)) > -1){
    xacml_init(); /* if you look at the code, you'll see this function is empty */
    if(xacml_server_init(&server, ees_xacml_authorize, NULL) == XACML_RESULT_SUCCESS){
      xacml_server_set_fd(server, fd);
      xacml_server_start(server);
      return EES_SUCCESS;
    }
  }
  return EES_FAILURE;
}

/**
 * Handles incoming signals
 */
void signal_handler(int sig){
  printf("Got signal\n");
  if(signal_is_being_handled == 0){
    signal_is_being_handled = 1;
    switch(sig){
      case SIGHUP:
        EEF_log(LOG_NOTICE, "Got SIGHUP - restart \n");
        ees_stop();
        if(ees_start() == EES_FAILURE){
          EEF_log(LOG_ERR, "Unable to restart EES. Exiting...\n");
          exit(EXIT_FAILURE);
        }
        break;
      case SIGTERM:
        EEF_log(LOG_NOTICE, "Got SIGTERM - shutting down\n");
        ees_stop();
        exit(EXIT_SUCCESS);
        break;
    }
  }
}

/**
 * Main
 */
int main (int argc, char* argv[]){
  struct sigaction sa;

	if(argc > 1){
		config_file_path = argv[1];
	} else {
		config_file_path = "setup/ees.conf.example";
	}

  /* catch hangup and term signals */
  sa.sa_handler = signal_handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;

  if(sigaction(SIGHUP, &sa, NULL) != 0){
    EEF_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }
  if(sigaction(SIGTERM, &sa, NULL) != 0){
    EEF_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }

#if ENABLE_DEBUG
    cgul_set_log_options(cgul_get_log_options() | CGUL_STDDEBUG);
#endif
  if(ees_start() == EES_SUCCESS){
#if !ENABLE_DEBUG
    EEF_daemonize();
#endif

    ees_loop();
    return EXIT_SUCCESS;
  }

  return EXIT_FAILURE;

}

